/* 
 * 	$Source: /home/fergia/CVS/tcfs-linux/tcfslib-1.0.0/src/tcfs_key.c,v $
 *	$Revision: 1.1.1.1 $
 * 	$Date: 2000/04/07 11:21:25 $
 *	$State: Exp $
 *      $Author: fergia $
 *      $Lockers$	
 */

#include <stdarg.h> /* For variable argument list functions */


#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <mntent.h>

#include "tcfs_global.h"
#include "tcfs_errno.h"
#include "tcfs_key.h"
#include <dlfcn.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>


/* #if defined (__GLIBC__)
#define _LINUX_SOCKET_H
#endif
#define _I386_BITOPS_H */
#include <linux/tcfs.h>
#include <linux/tcfs_ioctl.h>

/* This is initialized with the dynamic library's name */
static char libname[LINE]="\0"; 

static void* handle;
static void* (*init_key)(char*);
static char* (*encrypt)(char*,int,void*,int*);
static char* (*decrypt)(char*,int,void*,int*);
static void (*delete_key)(void*);



/* It fetch the dynamic library's name of the cipher to use. The result is put in libname. On success it return 0. It is also an internal function. */
static int _tcfs_get_library_name(void) {

	FILE* conf;
	char buf[LINE],lib[LINE],*temp;
	unsigned short state=0;

	tcfs_errno=OK;


#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_get_library_name: Execute _tcfs_get_library_name.\n");
#endif

	if ((conf=fopen(TCFS_CIPHER_CONF,"r"))==NULL)  
		return(tcfs_errno=ER_CONF_FILE);

		
	while (fgets(buf,LINE,conf)!=NULL) { 

		if (*buf!=TCFS_CIPHER_COMMENT && *buf!='\n') {
			
			temp=strtok(buf,DELIM);

			if (temp==NULL) continue; /* The line is empty */
			if (*temp==TCFS_CIPHER_COMMENT) continue;
	
			if (state==0) {
				if (strcmp(temp,TCFS_CIPHER_KEYWORD)==0) {
					temp=strtok(NULL,DELIM);
					if (temp==NULL) break; 
					strcpy(lib,temp);
					state=1;

#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_get_library_name: library: %s\n",lib);
#endif

					continue;	
				}
				else break;
			}
			if (state==1) {
				if (strcmp(temp,lib)==0) {
					temp=strtok(NULL,DELIM);
					if (temp==NULL) break; 
					strcpy(libname,temp);

#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_get_library_name: library name: %s\n",libname);
#endif
					break;
				}
			}
		}
	}
	if (strlen(libname)==0)  
		tcfs_errno=ER_BAD_CONF_FILE;
	
	
	fclose(conf);
	return (tcfs_errno);
}

/* It is used to load function from the dynamic library. On success it returns 0. It is also an internal function. */
static int _tcfs_load_cipher_function(void) {
	
#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_load_cipher_function: in tcfs_load_cipher_function.\n");
#endif

	handle=dlopen(libname,RTLD_NOW);

	if (handle==NULL) 
		return(tcfs_errno=ER_LOAD_LIBRARY);
	
#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_load_cipher_function: Loading init_key\n");
#endif

	init_key=(void* (*)(char*)) dlsym(handle,TCFS_INIT_KEY);
	if (dlerror()!=NULL) {
		dlclose(handle);
		return(tcfs_errno=ER_LOAD_FUNCTION);
	}

	encrypt=(char* (*)(char*, int, void*, int*)) dlsym(handle,TCFS_ENCRYPT);		
	if (dlerror()!=NULL) {
		dlclose(handle);
		return(tcfs_errno=ER_LOAD_FUNCTION);
	}

#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_load_cipher_function: loading %s\n",TCFS_DECRYPT);
#endif

	decrypt=(char* (*)(char*, int, void*, int*)) dlsym(handle,TCFS_DECRYPT);		
	if (dlerror()!=NULL) {
		dlclose(handle);
		return(tcfs_errno=ER_LOAD_FUNCTION);
	}
	
#ifdef TCFS_DEBUG
	fprintf(stderr,"_tcfs_load_cipher_function: loading %s.\n",TCFS_DELETE_KEY);
#endif

	delete_key=(void (*)(void*)) dlsym(handle,TCFS_DELETE_KEY);
	if (dlerror()!=NULL) {
		dlclose(handle);
		return(tcfs_errno=ER_LOAD_FUNCTION);
	}

	return(tcfs_errno=OK);
}

/* It is used to unload the dynamic library. It's an internal function.*/
static void _tcfs_unload_cipher_function(void) {

	dlclose(handle);

}
	
/* It is used to encrypt the tcfs key of an user. The encrypt's key is the user's password. On success it returns the encrypted tcfs_key. On error it returns NULL. Flag is used to distinguish the user key from the group key. Note that the returned string can be greater than the input, but it's a base 64 encoding string. So it has not the '\0' character. */
char* 
tcfs_encrypt_key (char* tcfs_key, char *password, int flag)
{

	void *ks;
	char *encrypted,*encoded;
	char pass[_PASSWORD_LEN];
	int keysize=(flag==TCFS_USER_KEY)?KEYSIZE:GROUPKEYSIZE,len;

#ifdef TCFS_DEBUG
	fprintf(stderr,"tcfs_encrypt_key: Execute tcfs_encrypt_key.\n");
	fprintf(stderr,"tcfs_encrypt_key: init_key %X\n",(unsigned int) *init_key);
	fprintf(stderr,"tcfs_encrypt_key: Value of handle %p\n",handle);
	fprintf(stderr,"tcfs_encrypt_key: Value of libname %s\n",libname);
#endif

	tcfs_errno=OK;

	if (strlen(libname)==0) { 
		if (_tcfs_get_library_name()!=0) return (char*)0;
  
#ifdef TCFS_DEBUG
	fprintf(stderr,"tcfs_encrypt_key: _tcfs_get_library_name return success.\n");
#endif
	
	}

	if (_tcfs_load_cipher_function()!=0) return (char*)0;
	
	strcpy (pass, password);
  	while (strlen(pass)<KEYSIZE)
      		strcat (pass, password);
   	
#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_encrypt_key: Before init_key pass=%s, strlen(pass)=%d",pass,strlen(pass));
	fprintf(stderr,"\ntcfs_encrypt_key: init_key %X",(unsigned int) *init_key);
#endif

	ks=init_key(pass);
	encrypted=encrypt(tcfs_key,keysize,ks,&len);
	delete_key(ks);

#ifdef TCFS_DEBUG
	{
	int i=0;
	fprintf(stderr,"\ntcfs_encrypt_key: tcfs_key crypted = ");
	for (;i<len;i++)
		fprintf(stderr,"%X ",(unsigned char) encrypted[i]);
	fprintf(stderr,"\n");
	fprintf(stderr,"\ntcfs_encrypt_key: Value of handle %p",handle);
	}
#endif
	
	_tcfs_unload_cipher_function();		

  	encoded=(char*) tcfs_base64_encode (encrypted,len);

	return(encoded);
}


/* It is used to decrypt the tcfs key of an user. The decrypt's key is the user's password. On success it returns the decrypted tcfs_key. On error it returns NULL. Flag is used to distinguish the user key from the group key. Note that the returned string must be keysize long. */
char * 
tcfs_decrypt_key (char *tcfs_key, char *password, int flag)
{
  	char pass[_PASSWORD_LEN];
	char *decrypted,*decoded;
	void *ks;
	int len,keysize=(flag==TCFS_USER_KEY)?KEYSIZE:GROUPKEYSIZE,decryptlen;

#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_decrypt_key: Execute tcfs_decrypt_key");
	fprintf(stderr,"\ntcfs_decrypt_key: Value of libname %s",libname);
#endif

	tcfs_errno=OK;

	if (strlen(libname)==0) { 
		if (_tcfs_get_library_name()!=0) return (char*)0;
  
#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_decrypt_key: _tcfs_get_library_name return success.");
#endif
	
	}
	
	if (_tcfs_load_cipher_function()!=0) return (char*)0;
	

#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_decrypt_key: before tcfs_base64_decode, value of tcfs_key: %s",tcfs_key);
#endif

	if ((decoded = (char*) tcfs_base64_decode (tcfs_key,&len))==NULL) {
		_tcfs_unload_cipher_function();	
		return (char*) 0;
	}
	
	strcpy (pass, password);
  	while (strlen(pass)<KEYSIZE)
      		strcat (pass, password);

	ks=init_key(pass);
	/* On success, decryptlen must ne egual to keysize */
	decrypted=decrypt(decoded,len,ks,&decryptlen);
	delete_key(ks);
		
#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_decrypt_key: Len of the decrypted string: %d\n",decryptlen);
#endif

	_tcfs_unload_cipher_function();	
	free(decoded);

	if (decryptlen!=keysize) {
		tcfs_errno=ER_TCFS_DECRYPT;
		return (char*)0;
	}

	return(decrypted);
}



/* It's the interface to the ioctl function. It allows the call of the ioctl function on all the tcfs filesystems. On success it returns 0. It's an internal function. */

static int 
_tcfs_ioctl_interface(int function, char *arg, char *filesystem, char *filename)
{
	FILE *mtab;
	int fd,counter=0,ioctl_success=0;
	struct mntent *mp;

	tcfs_errno=OK;
	if ((mtab = setmntent (TCFS_MTAB_PATH, "r")) == NULL) 
		return (tcfs_errno=ER_MTAB_OPEN);

	while ((mp = getmntent (mtab)) != NULL)
	{
		if (strcmp (mp->mnt_type, "tcfs")==0)
		{
			counter++;
#ifdef TCFS_DEBUG 
	fprintf(stderr,"_tcfs_ioctl_interface: filesystem to process: %s\n",mp->mnt_dir);
#endif

			if ((fd = open (mp->mnt_dir, O_RDONLY)) < 0)
			{
				endmntent(mtab);
				return(tcfs_errno=ER_TCFS_OPEN);
			}
			else
			{

#ifdef TCFS_DEBUG 
	fprintf(stderr,"_tcfs_ioctl_interface: before ioctl\n");
	fprintf(stderr,"TCFS_IOC_LOGIN=%d\n", TCFS_IOC_LOGIN);
	fprintf(stderr,"TCFS_IOC_LOGIN: %d\n", function);
	fprintf(stderr,"uid_t: %d\n", sizeof(uid_t));
#endif
				if (ioctl (fd, function, arg)>=0)
					ioctl_success++; 
			

#ifdef TCFS_DEBUG 
	fprintf(stderr,"_tcfs_ioctl_interface: after ioctl\n");
	fprintf(stderr,"_tcfs_ioctl_interface: ioctl_success %d\n", ioctl_success);
#endif

			}
			close (fd);
		}
	}
	endmntent (mtab);
	
	if (counter==0) tcfs_errno=ER_NO_TCFS_MOUNTED;
	if (ioctl_success==0) tcfs_errno=ER_IOCTL;

	return tcfs_errno;
}

/* This passes the tcfs_key to the kernel.
*/
int 
tcfs_enable (int uid, char *tcfskey)
{
  struct login user;
	
  tcfs_errno=OK;	

  memset (&user, 0, sizeof(struct login));
  user.uid=uid;	
  memcpy (user.deskey, (unsigned char*) tcfskey, KEYSIZE);
  memcpy (user.ciphername, "default\0", 8);

  if (_tcfs_ioctl_interface(TCFS_IOC_LOGIN, (char *) &user, NULL, NULL)!=0)
	if (tcfs_errno==ER_IOCTL) 
		tcfs_errno=ER_ENABLE;

  return(tcfs_errno);	
	
}

/* It removes the tcfs_key from the kernel; user_name is used to get the uid, while flag is used to remove all the session's keys. If flag is TCFS_COUNTER_ONE, only the current sessions' key is removed, otherwise all the sessions.s key are removed. */

int tcfs_disable(int uid,int flag) {


	tcfs_errno=OK;
	
  	if (_tcfs_ioctl_interface((flag==TCFS_COUNTER_ALL)?TCFS_IOC_FULLOUT:TCFS_IOC_LOGOUT, (char *) &uid, NULL, NULL)!=0)
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_DISABLE;

	return (tcfs_errno);
}
	
/* It passes the tcfs group key to the kernel. The user_name is used to get the uid, the group name is used to get the group id, part is the number of partecipants and threshold is the minimum number of user to get the key. The key to pass to the kernel is tcfskey. On success it returns 0. */
int 
tcfs_group_enable(int uid, int gid, int part, int threshold, char *tcfskey)
{
  struct tgroup group;
	
  tcfs_errno=OK;	


  group.gid = gid;	
  group.uid = uid;
  group.n=part;
  group.soglia=threshold;
  memcpy (group.deskey, (unsigned char*) tcfskey, GROUPKEYSIZE);

#ifdef TCFS_DEBUG
	{
		int i;
		fprintf(stderr,"tcfs_group_enable: passing...\n\tkey:");
		for (i=0;i<GROUPKEYSIZE;i++) 
			fprintf(stderr,":%X",group.deskey[i]);
		fprintf(stderr,"\n");
		fprintf(stderr,"gid %d uid %d part %d threshold %d\n",gid,uid,part,threshold);
		
	}
#endif

  if (_tcfs_ioctl_interface(TCFS_IOC_GLOGIN, (char *) &group, NULL, NULL)!=0)
	if (tcfs_errno==ER_IOCTL) 
		tcfs_errno=ER_ENABLE;

  return(tcfs_errno);	
	
}

/* It removes the tcfs group key from the kernel. The user_name is used to get the uid, the group name is uset to get the group id. On success it returns 0.
Flag is used to remove all the session's keys. If flag is TCFS_COUNTER_ONE, only the current session's key is removed, otherwise all the session's keys are removed. By default flag is TCFS_COUNTER_ONE. With TCFS_COUNTER_ONE set, the user must specify user_name, part and threshold. With TCFS_COUNTER_ALL set, the user doesn't need to specify the user_name and so on.   */
/*int 
tcfs_group_disable(int uid, int gid, int part, int threshold, int flag)
{
		
  tcfs_errno=OK;	
	

  if (flag==TCFS_COUNTER_ALL) {

	
  	if (_tcfs_ioctl_interface(TCFS_IOC_GFULLOUT, (char *) &gid, NULL, NULL)!=0)
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_DISABLE;
  }
  else {
	struct tgroup group;

	group.gid = gid;	
	group.uid = uid;
	group.n=part;
	group.soglia=threshold;
  	
	if (_tcfs_ioctl_interface(TCFS_IOC_GLOGOUT, (char *) &group, NULL, NULL)!=0)
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_DISABLE;
  }
  
  return(tcfs_errno);	
	
}*/

int 
tcfs_group_disable(int gid, int flag, ...)
{
		
  tcfs_errno=OK;	
	

  if (flag==TCFS_COUNTER_ALL) {

 
#ifdef TCFS_DEBUG
	fprintf(stderr,"tcfs_group_disable: gid:%d\n",gid);
#endif
 
  	if (_tcfs_ioctl_interface(TCFS_IOC_GFULLOUT, (char *) &gid, NULL, NULL)!=0)
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_DISABLE;
  }
  else {
	va_list list;
	struct tgroup group;
	int uid, part, threshold;
	
#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_group_disable: Before the scanning of arguments");
#endif

	/* Scanning the arguments */	
	va_start(list,flag);
	uid=va_arg(list,uid_t);
	part=va_arg(list,int);
	threshold=va_arg(list,int);
	va_end(list);

#ifdef TCFS_DEBUG
	fprintf(stderr,"\ntcfs_group_disable: gid:%d user_id:%d, part:%d, threshold:%d",gid,uid,part,threshold);
#endif

	group.gid = gid;	
	group.uid = uid;
	group.n=part;
	group.soglia=threshold;
  	
	if (_tcfs_ioctl_interface(TCFS_IOC_GLOGOUT, (char *) &group, NULL, NULL)!=0)
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_DISABLE;
  }
  
  return(tcfs_errno);	
	
}

/* It is used to set or unset the permanent flag. If flag is TCFS_SET_FLAG, then the flag is set, if flag is TCFS_DEL_FLAG, the flag is unset. */
int
tcfs_permanent (int flag)
{
	tcfs_errno=OK;
	
	if (_tcfs_ioctl_interface((flag==TCFS_SET_FLAG)?TCFS_IOC_SETFIX:TCFS_IOC_DELFIX, /*(char *)&uid*/ NULL, NULL, NULL)!=0)
                if (tcfs_errno==ER_IOCTL)
                        tcfs_errno=ER_PERM;

	return (tcfs_errno);
}

/* It is used to set/unset the permanent flag of a piece of key that forms the group key. 
If flag is TCFS_SET_FLAG, then the flag is set, if flag is TCFS_DEL_FLAG, the flag is unset. */
int
tcfs_group_permanent (int gid, int flag)
{


  	if (_tcfs_ioctl_interface((flag==TCFS_SET_FLAG)?TCFS_IOC_GSETFIX:TCFS_IOC_GDELFIX, (char *) &gid, NULL, NULL)!=0)
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_PERM;

	return (tcfs_errno);
}

/* It gives back the counter of the key. On error it returns -1, else it returns the counter's value. */
int
tcfs_getcount (void)
{
	unsigned int count=0;

	tcfs_errno=OK;

  	if (_tcfs_ioctl_interface(TCFS_IOC_GETCOUNT, (char *) &count, NULL, NULL)!=0) {
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_COUNT;
		return -1;
	}
	else 
		return (int) count;
}


/* It gives back the counter of groupkey's piece. On error it returns -1, else it returns the counter's value. */
int
tcfs_group_getcount (int gid)
{
	unsigned int arg;

	tcfs_errno=OK;


	arg = (unsigned int) gid;
  	
	if (_tcfs_ioctl_interface(TCFS_IOC_GGETCOUNT, (char *) &arg, NULL, NULL)!=0) {
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_COUNT;
		return -1;
	}
	else 
		return (int) arg;
}


/* It gives back the permanent flag's value of user key. On error it returns -1, else it returns the correct value. */
int
tcfs_getpermanent (void)
{	
	unsigned char permanent;

	tcfs_errno=OK;

  	if (_tcfs_ioctl_interface(TCFS_IOC_GETFIX, (char *) &permanent, NULL, NULL)!=0) {
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_PERM;
		return -1;
	}
	else 
		return (int) permanent;
	
}

/* It gives back the permanent flag's value of groupkey's piece. On error it returns -1, else it returns the correct value. */
int
tcfs_group_getpermanent (int gid)
{
	unsigned int arg;

	tcfs_errno=OK;

	arg = (unsigned int) gid;
  	
	if (_tcfs_ioctl_interface(TCFS_IOC_GGETFIX, (char *) &arg, NULL, NULL)!=0) {
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_PERM;
		return -1;
	}
	else 
		return (int) arg;
}

/* It gives back informatione about the tcfs version used. It gives the version number, the minor version number and the cipher. It creates memory for cipher. So the user must free it. On success it returns 0. */ 
int tcfs_getversion(short *version, short* minor, char** cipher) 
{   
	union
	{ 
		long arg_long; 
		char arg_char[sizeof(long)];
	} arg;   
	int i; 

	tcfs_errno=OK;

	if (_tcfs_ioctl_interface(TCFS_IOC_GETVERSION, (char *) &arg, NULL, NULL)!=0) {
		if (tcfs_errno==ER_IOCTL) 
			tcfs_errno=ER_VERSION;

		return(tcfs_errno);
	}

	*version=(short) arg.arg_char[0];
	*minor=(short) arg.arg_char[1];
	i=arg.arg_char[3];

#ifdef TCFS_DEBUG
	fprintf(stderr,"tcfs_getversion: value of i: %d \n",i);
#endif

	*cipher=malloc(LINE);
	(i>=NUM_CIPHER)? strcpy(*cipher,"Unknown cipher"):strcpy(*cipher,tcfs_engines[i]);

	return(tcfs_errno);
}
